#
#  Plex Extension Framework
#  Copyright (C) 2008-2010 Plex Development Team (James Clarke, Elan Feingold). All Rights Reserved.
#

import Framework

class LocaleKit(Framework.bases.BaseKit):
  def _init(self):
    self._globals = dict(
      L = self.LocalString,
      F = self.LocalStringWithFormat,
    )
    self.Language = self._core.localization.language
    
  @property
  def DefaultLocale(self):
    """
      Returns the default locale currently in use by the plug-in, e.g. ``en-us``.
    """
    return self._core.localization.default_locale
    
  @DefaultLocale.setter
  def DefaultLocale(self, value):
    self._core.localization.default_locale = value
    
  @property
  def Geolocation(self):
    """
      Returns the user's country, obtained via IP-based geolocation, e.g. ``US``.
    """
    return self._core.networking.http_request("http://geo.plexapp.com/geolocate.php", cacheTime=86400).content
  
  @property
  def CurrentLocale(self):
    """
      Returns the locale of the user currently making a request to the plug-in, or *None* if no locale was provided.
    """
    if self._context and Framework.constants.header.language in self._context.headers:
      return self._context.headers[Framework.constants.header.language]
    return None
  
  def LocalString(self, key):
    """
      Retrieves the localized version of a string with the given key. Strings from the user's
      current locale are given highest priority, followed by strings from the default locale.
      
      See `String files` for more information on proividing localized versions of strings.
    """
    return self._core.localization.local_string(key, self.CurrentLocale)
    
  def LocalStringWithFormat(self, key, *args):
    """
      Retrieves the localized version of a string with the given key, and formats it using the
      given arguments.
    """
    return self._core.localization.local_string_with_format(key, self.CurrentLocale, *args)
    
  def _requres_context(self, context):
    return Framework.constants.header.language in self._context.headers