#
#  Plex Extension Framework
#  Copyright (C) 2008-2010 Plex Development Team (James Clarke, Elan Feingold). All Rights Reserved.
#

import Framework
import mimetypes


class ResourceKit(Framework.bases.BaseKit):

  def _init(self):
    self._globals = dict(
      R = self.ExternalPath,
      S = self.SharedExternalPath,
    )
    
  def ExternalPath(self, itemname):
    """
      Generates an externally accessible path for the named resource. This method is usually used to
      assign bundled images to object attributes.
      
      .. note:: The alias :func:`R()` is equivalent to this method.
    """
    return self._core.runtime.external_resource_path(itemname)
  
  def SharedExternalPath(self, itemname):
    """
      Generates an externally accessible path for the named shared resource. A list of valid
      shared resource names is provided below.
      
      .. note:: The alias :func:`S()` is equivalent to this method.
    """
    return self._core.runtime.external_shared_resource_path(itemname)  
  
  def Load(self, itemname, binary=True):
    """
      Loads the named resource file and returns the data as a string.
    """
    return self._core.storage.load_resource(itemname, binary)
    
  def LoadShared(self, itemname, binary=True):
    """
      Loads the named shared resource file and returns the data as a string.
    """
    return self._core.storage.load_shared_resource(itemname, binary)
    
  def GuessMimeType(self, path):
    """
      Attempts to guess the MIME type of a given path based on its extension.
    """
    return mimetypes.guess_type(path, strict=False)[0]
    
  def AddMimeType(self, mimetype, extension):
    """
      Assigns the specified MIME type to the given extension.
    """
    if extension[0] != '.': extension = '.' + extension
    mimetypes.add_type(mimetype, extension)