from systemservice import SystemService

class CodeService(SystemService):
  def __init__(self, system):
    SystemService.__init__(self, system)
    Route.Connect('/system/services/url/lookup', self.cs_url_lookup)
    Route.Connect('/system/services/search', self.cs_search)
    Route.Connect('/system/services/relatedcontent/lookup', self.cs_relatedcontent_lookup)
    
    if hasattr(Core.config, 'daemonized') and Core.config.daemonized:
      Route.Connect('/services/url/lookup', self.cs_url_lookup)
      Route.Connect('/services/search', self.cs_search)
      Route.Connect('/services/relatedcontent/lookup', self.cs_relatedcontent_lookup)
      
  def cs_url_lookup(self, url, limit=None):
    Log.Debug("Looking up URL '%s'", url)

    # Get the metadata object for the given URL
    c = URLService.MetadataObjectForURL(url, allow_deferred=True, in_container=True)
    # If an object was returned, add it to a container and return it
    if c != None:
      obj = c.objects[0]
      
      # Move the user agent and http cookies (if set) to the container instead of the object
      try:
        if obj.user_agent:
          c.user_agent = obj.user_agent
          obj.user_agent = None
      except:
        pass
        
      try:
        if obj.http_cookies:
          c.http_cookies = obj.http_cookies
          obj.http_cookies = None
      except:
        pass
        
      if limit and len(obj) > int(limit):
        obj.items = obj.items[:int(limit)]
        
      return c

  def cs_search(self, query, identifier, name=None):
    result = SearchService.Query(query, identifier, name)
    return result
    
  def cs_relatedcontent_lookup(self, url):
    obj = URLService.MetadataObjectForURL(url)
    
    if obj == None:
      return None
      
    identifier = URLService.ServiceIdentifierForURL(url)
    if identifier == None:
      return None
      
    result = RelatedContentService.RelatedContentForObject(obj)
    return result
